#include "sim_diff.h"
#include <ctype.h>

/* TODO: Optimise reads to not read whole block and then overwrite. */

#define MAX_DIFFERENCE_FILES 16
#define DIFF_RESERVED_HEADER_BYTES 1024

/* debugging bitmaps */
#define DBG_TRC  0x0001                                 /* trace routine calls */
#define DBG_WRN  0x0002                                 /* display warnings */
#define DBG_INF  0x0004                                 /* display informational messages (high level trace) */


struct difference_header
{
	t_addr VirtualDiskSize;
	uint32 BlockSize;
	t_addr NumBlocks;
	uint8  Reserved[DIFF_RESERVED_HEADER_BYTES];
	uint8 *Buffer; /* used to store a block in memory while it is worked on, not actually part of the header on disk */
	t_addr BlockIndex[]; /* memory for this is malloced on to the end of the header */
};

typedef struct difference_header DIFFERENCE_HEADER;

struct difference_file
{
	int InUse;
	UNIT *pUnit;
	FILE *DifferenceFile;
	DIFFERENCE_HEADER *pHeader;
	char *FileName;
};

typedef struct difference_file DIFFERENCE_FILE;

DIFFERENCE_FILE DifferenceFileTable[MAX_DIFFERENCE_FILES];

DIFFERENCE_FILE *diff_find_difference_file_for_unit(UNIT *uptr);
DIFFERENCE_FILE *diff_find_difference_file_for_file(FILE *file);
t_stat diff_allocate_difference_file_table_slot(UNIT *uptr, DIFFERENCE_FILE **ppDifferenceFile);
void diff_deallocate_difference_file_table_slot(DIFFERENCE_FILE *pDifferenceFile);
t_stat diff_open_difference_file(DIFFERENCE_FILE *pDifferenceFile, uint32 blockSize);
t_stat diff_create_difference_file(DIFFERENCE_FILE *pDifferenceFile, char *name, uint32 blockSize);
void diff_close_difference_file(DIFFERENCE_FILE *pDifferenceFile);
void diff_delete_difference_file(DIFFERENCE_FILE *pDifferenceFile);
t_stat diff_merge_difference_file(DIFFERENCE_FILE *pDifferenceFile);
t_addr diff_calculate_num_blocks(DIFFERENCE_FILE *pDifferenceFile, uint32 blockSize);
t_addr diff_get_block_index_table_position();
t_addr diff_get_block_index_table_entry_position(t_addr blockNumber);
t_addr diff_calculate_block_number_for_address(DIFFERENCE_FILE *pDifferenceFile, t_addr address);
void diff_create_header_and_index(DIFFERENCE_FILE *pDifferenceFile, uint32 blockSize, t_addr numBlocks);
void diff_delete_header_and_index(DIFFERENCE_FILE *pDifferenceFile);
t_stat diff_write_header(DIFFERENCE_FILE *pDifferenceFile);
t_stat diff_read_header(DIFFERENCE_FILE *pDifferenceFile);
t_stat diff_update_block_index_entry(DIFFERENCE_FILE *pDifferenceFile, t_addr blockNumber, t_addr newValue);
t_stat diff_write_block(DIFFERENCE_FILE *pDifferenceFile, t_addr blockNumber, uint8 *pBuffer, uint32 blockOffset, size_t byteCount);
t_stat diff_update_existing_block(DIFFERENCE_FILE *pDifferenceFile, t_addr blockNumber, uint8 *pBuffer, uint32 blockOffset, size_t byteCount);
t_stat diff_create_new_block(DIFFERENCE_FILE *pDifferenceFile, t_addr blockNumber);
t_stat diff_check_block_number(DIFFERENCE_FILE *pDifferenceFile, t_addr blockNumber);
t_stat diff_read_block(DIFFERENCE_FILE *pDifferenceFile, t_addr blockNumber, uint8 *pBuffer, uint32 blockOffset, uint32 byteCount);
t_stat diff_read_diff_block(DIFFERENCE_FILE *pDifferenceFile, t_addr blockNumber, uint8 *pBuffer, uint32 blockOffset, uint32 byteCount);
t_stat diff_read_original_block(DIFFERENCE_FILE *pDifferenceFile, t_addr blockNumber);
t_stat diff_move(FILE *fptr, t_addr position);
t_stat diff_move_eof(FILE *fptr);
void diff_show_stats(DIFFERENCE_FILE *pDifferenceFile);
t_stat diff_indirect_fwrite(void *bptr, size_t size, size_t count, FILE *fptr);
t_stat diff_indirect_fread(void *bptr, size_t size, size_t count, FILE *fptr);
t_stat diff_direct_fwrite(void *bptr, size_t size, size_t count, FILE *fptr);
t_stat diff_direct_fread(void *bptr, size_t size, size_t count, FILE *fptr);
t_stat diff_fflush(FILE *fptr);

extern FILE *sim_log;
extern FILE *sim_deb;  

#define TRACE_BYTES_PER_LINE 64

void diff_trace_block_line(uint8 *buf, int length, char *prefix)
{
	char hex[TRACE_BYTES_PER_LINE*3+1];
	char ascii[TRACE_BYTES_PER_LINE+1];
	/*char ts[9];*/
	/*time_t t;*/
	int i;
	hex[0] = 0;
	ascii[TRACE_BYTES_PER_LINE] = 0;

	/*t = time(NULL);*/
	/*strftime(ts, 9, "%H:%M:%S", localtime(&t));*/
	for (i = 0; i<TRACE_BYTES_PER_LINE; i++)
	{
		if (i>=length)
		{
			strcat(hex, "   ");
			ascii[i] = ' ';
		}
		else
		{
		    char hexByte[4];
			sprintf(hexByte, "%02X ", buf[i]);
			strcat(hex, hexByte);
			if (isprint(buf[i]))
			{
				ascii[i] = (char)buf[i];
			}
			else
			{
				ascii[i] = '.';
			}
		}
	}

	fprintf(sim_deb, "%s %s  %s\r\n", prefix, hex, ascii);
}

void diff_trace_block(uint8 *buf, int length, char *prefix)
{
	int i;
	if (length >= 0)
	{
		for(i = 0; i < length / TRACE_BYTES_PER_LINE; i++)
		{
			diff_trace_block_line(buf + i*TRACE_BYTES_PER_LINE, TRACE_BYTES_PER_LINE, prefix);
		}

		if (length %TRACE_BYTES_PER_LINE > 0)
		{
			diff_trace_block_line(buf + length/TRACE_BYTES_PER_LINE, length % TRACE_BYTES_PER_LINE, prefix);
		}
	}
}

DIFFERENCE_FILE *diff_find_difference_file_for_unit(UNIT *uptr)
{
	int i;
	DIFFERENCE_FILE *ans = NULL;
	for (i = 0; i < MAX_DIFFERENCE_FILES; i++)
	{
		if (DifferenceFileTable[i].InUse && DifferenceFileTable[i].pUnit->fileref == uptr->fileref)
		{
			ans = &DifferenceFileTable[i];
			break;
		}
	}

	return ans;
}

DIFFERENCE_FILE *diff_find_difference_file_for_file(FILE *file)
{
	int i;
	DIFFERENCE_FILE *ans = NULL;
	for (i = 0; i < MAX_DIFFERENCE_FILES; i++)
	{
		if (DifferenceFileTable[i].InUse && DifferenceFileTable[i].pUnit->fileref == file)
		{
			ans = &DifferenceFileTable[i];
			break;
		}
	}

	return ans;
}

t_stat diff_allocate_difference_file_table_slot(UNIT *uptr, DIFFERENCE_FILE **ppDifferenceFile)
{
	t_stat ans = SCPE_OK;
	int i;
	int index = -1;
	*ppDifferenceFile = NULL;

	for (i = 0; i < MAX_DIFFERENCE_FILES; i++)
	{
		if (!DifferenceFileTable[i].InUse)
		{
			index = i;
			break;
		}
	}

	if (index >= 0)
	{
		DifferenceFileTable[index].InUse = TRUE;
		DifferenceFileTable[index].pUnit = uptr;
		*ppDifferenceFile = &DifferenceFileTable[index];
	}
	else
	{
		ans = SCPE_NOMOREDIFF;
	}

	return ans;
}

void diff_deallocate_difference_file_table_slot(DIFFERENCE_FILE *pDifferenceFile)
{
	pDifferenceFile->InUse = FALSE;
}

t_stat diff_open_difference_file(DIFFERENCE_FILE *pDifferenceFile, uint32 blockSize)
{
	t_stat ans = SCPE_OK;
	pDifferenceFile->DifferenceFile = sim_fopen(pDifferenceFile->FileName, "rb+");
	if (pDifferenceFile->DifferenceFile == NULL)
	{
		if (errno == ENOENT)
		{
			ans = diff_create_difference_file(pDifferenceFile, pDifferenceFile->FileName, blockSize);
		}
		else
		{
			printf("Failed to open %s. Error %d\n", pDifferenceFile->FileName, errno);
			ans = SCPE_OPENERR;
		}
	}
	else
	{
		printf("Opened difference file %s\n", pDifferenceFile->FileName);
		ans = diff_read_header(pDifferenceFile);
	}

	return ans;
}

t_stat diff_create_difference_file(DIFFERENCE_FILE *pDifferenceFile, char *name, uint32 blockSize)
{
	t_stat ans = SCPE_OK;
	pDifferenceFile->DifferenceFile = sim_fopen(name, "wb+");
	if (pDifferenceFile->DifferenceFile == NULL)
	{
		printf("Failed to create %s. Error %d\n", name, errno);
		ans = SCPE_OPENERR;
	}
	else
	{
		t_addr numBlocks;
		printf("Created difference file %s\n", name);
		numBlocks = diff_calculate_num_blocks(pDifferenceFile, blockSize);
		diff_create_header_and_index(pDifferenceFile, blockSize, numBlocks);
		ans = diff_write_header(pDifferenceFile);
	}

	return ans;
}

void diff_close_difference_file(DIFFERENCE_FILE *pDifferenceFile)
{
	diff_delete_header_and_index(pDifferenceFile);
	fclose(pDifferenceFile->DifferenceFile);
}

void diff_delete_difference_file(DIFFERENCE_FILE *pDifferenceFile)
{
	remove(pDifferenceFile->FileName);
}

t_addr diff_calculate_num_blocks(DIFFERENCE_FILE *pDifferenceFile, uint32 blockSize)
{
	t_addr ans = pDifferenceFile->pUnit->capac / blockSize;
	if ((pDifferenceFile->pUnit->capac % blockSize) > 0) ans ++;

	printf("Capacity %I64u\n", pDifferenceFile->pUnit->capac);
	printf("BlockSize %u\n", blockSize);
	printf("Num Blocks %I64u\n", ans);
	return ans;
}

void diff_create_header_and_index(DIFFERENCE_FILE *pDifferenceFile, uint32 blockSize, t_addr numBlocks)
{
	uint32 i;
	size_t headerSize = sizeof(DIFFERENCE_HEADER) + (size_t)numBlocks * sizeof(t_addr); /* cast on NumBlocks will cause failure on very large files, needs 64-bit target */
	pDifferenceFile->pHeader = malloc(headerSize);
	pDifferenceFile->pHeader->VirtualDiskSize = pDifferenceFile->pUnit->capac;
	pDifferenceFile->pHeader->BlockSize = blockSize;
	pDifferenceFile->pHeader->NumBlocks = numBlocks;
	for (i = 0; i<numBlocks; i++)
	{
		pDifferenceFile->pHeader->BlockIndex[i] = 0;
	}

	pDifferenceFile->pHeader->Buffer = malloc(blockSize);
}

void diff_delete_header_and_index(DIFFERENCE_FILE *pDifferenceFile)
{
	free(pDifferenceFile->pHeader->Buffer);
	free(pDifferenceFile->pHeader);
	pDifferenceFile->pHeader = NULL;
}

t_stat diff_write_header(DIFFERENCE_FILE *pDifferenceFile)
{
	t_stat ans = SCPE_OK;

	ans = diff_move(pDifferenceFile->DifferenceFile, 0);

	if (ans == SCPE_OK) ans = diff_indirect_fwrite(&pDifferenceFile->pHeader->VirtualDiskSize, sizeof(pDifferenceFile->pHeader->VirtualDiskSize), 1, pDifferenceFile->DifferenceFile);

	if (ans == SCPE_OK) ans = diff_indirect_fwrite(&pDifferenceFile->pHeader->BlockSize, sizeof(pDifferenceFile->pHeader->BlockSize), 1, pDifferenceFile->DifferenceFile);

	if (ans == SCPE_OK) ans = diff_indirect_fwrite(&pDifferenceFile->pHeader->NumBlocks, sizeof(pDifferenceFile->pHeader->NumBlocks), 1, pDifferenceFile->DifferenceFile);

	if (ans == SCPE_OK) ans = diff_indirect_fwrite(&pDifferenceFile->pHeader->Reserved, sizeof(uint8), DIFF_RESERVED_HEADER_BYTES, pDifferenceFile->DifferenceFile);

	if (ans == SCPE_OK) ans = diff_indirect_fwrite(&pDifferenceFile->pHeader->BlockIndex, sizeof(t_addr), (size_t)pDifferenceFile->pHeader->NumBlocks, pDifferenceFile->DifferenceFile);  /* cast on NumBlocks will cause failure on very large files, needs 64-bit target */

	ans = diff_fflush(pDifferenceFile->DifferenceFile);

	return ans;
}

t_stat diff_read_header(DIFFERENCE_FILE *pDifferenceFile)
{
	t_stat ans = SCPE_OK;
	t_addr virtualDiskSize;
	uint32 blockSize;
	t_addr numBlocks;
	uint8 reserved[DIFF_RESERVED_HEADER_BYTES];

    ans = diff_move(pDifferenceFile->DifferenceFile, 0);

	if (ans == SCPE_OK) ans = diff_indirect_fread(&virtualDiskSize, sizeof(virtualDiskSize), 1, pDifferenceFile->DifferenceFile);

	if (ans == SCPE_OK) ans = diff_indirect_fread(&blockSize, sizeof(blockSize), 1, pDifferenceFile->DifferenceFile);

	if (ans == SCPE_OK) ans = diff_indirect_fread(&numBlocks, sizeof(numBlocks), 1, pDifferenceFile->DifferenceFile);

	if (ans == SCPE_OK) ans = diff_indirect_fread(&reserved, sizeof(uint8), DIFF_RESERVED_HEADER_BYTES, pDifferenceFile->DifferenceFile);

	diff_create_header_and_index(pDifferenceFile, blockSize, numBlocks);

	if (ans == SCPE_OK) ans = diff_indirect_fread(&pDifferenceFile->pHeader->BlockIndex, sizeof(t_addr), (size_t)pDifferenceFile->pHeader->NumBlocks, pDifferenceFile->DifferenceFile);  /* cast on NumBlocks will cause failure on very large files, needs 64-bit target */

	return ans;
}

t_addr diff_get_block_index_table_position()
{
	uint32 ans = sizeof(t_addr)
		+
		sizeof(uint32)
		+
		sizeof(t_addr)
		+
		sizeof(uint8) * DIFF_RESERVED_HEADER_BYTES;

	return ans;
}

t_addr diff_get_block_index_table_entry_position(t_addr blockNumber)
{
	return diff_get_block_index_table_position() + blockNumber * sizeof(t_addr);
}

t_stat diff_update_block_index_entry(DIFFERENCE_FILE *pDifferenceFile, t_addr blockNumber, t_addr newValue)
{
	t_stat ans = diff_check_block_number(pDifferenceFile, blockNumber);

	if (ans == SCPE_OK)
	{
		ans = diff_move(pDifferenceFile->DifferenceFile, diff_get_block_index_table_entry_position(blockNumber));

		if (ans == SCPE_OK)
		{
			if (ans == SCPE_OK) ans = diff_indirect_fwrite(&newValue, sizeof(newValue), 1, pDifferenceFile->DifferenceFile);
			if (ans == SCPE_OK)
			{
				pDifferenceFile->pHeader->BlockIndex[blockNumber] = newValue;
			}
		}
	}

	return ans;
}

t_addr diff_calculate_block_number_for_address(DIFFERENCE_FILE *pDifferenceFile, t_addr address)
{
	t_addr ans;

	ans = address / pDifferenceFile->pHeader->BlockSize;

	return ans;
}

/* Assumes data has already been flipped */
size_t diff_fwrite(void *bptr, size_t size, size_t count, FILE *fptr)
{
	size_t ans;
	size_t bytesWritten = 0;
	t_stat stat = SCPE_OK;
	DIFFERENCE_FILE *df = diff_find_difference_file_for_file(fptr);
	if (df != NULL)
	{
		t_addr startAddress = ftell(fptr);
		uint32 currentBlockOffset = startAddress % df->pHeader->BlockSize; /* offset into the current block */
		size_t bufferLength = size * count;
		t_addr startBlock = diff_calculate_block_number_for_address(df, startAddress);
		t_addr endBlock = diff_calculate_block_number_for_address(df, startAddress + bufferLength - 1);
		t_addr i;

		/*fprintf(sim_deb,"Start buffer write at %I64u, length = %u\r\n", startAddress, bufferLength);*/

		for (i = startBlock; i <= endBlock && stat == SCPE_OK; i++)
		{
			size_t bytesToWrite;
			bytesToWrite = df->pHeader->BlockSize - currentBlockOffset % df->pHeader->BlockSize;
			if (bytesToWrite > bufferLength - bytesWritten)
			{
				bytesToWrite = bufferLength - bytesWritten;
			}
			
			stat = diff_write_block(df, i, (uint8 *)bptr + bytesWritten, currentBlockOffset, bytesToWrite);
			
			if (stat == SCPE_OK)
			{
				bytesWritten += bytesToWrite;
				currentBlockOffset = (startAddress + bytesWritten) % df->pHeader->BlockSize;
			}
		}

		ans = bytesWritten / count;

		diff_move(fptr, startAddress + bufferLength);

	}
	else
	{
		ans = fwrite(bptr, size, count, fptr);
	}

	return ans;
}

t_stat diff_write_block(DIFFERENCE_FILE *pDifferenceFile, t_addr blockNumber, uint8 *pBuffer, uint32 blockOffset, size_t byteCount)
{
	t_stat ans = diff_check_block_number(pDifferenceFile, blockNumber);

	if (ans == SCPE_OK)
	{
		if (pDifferenceFile->pHeader->BlockIndex[blockNumber] == 0)
		{
			ans = diff_create_new_block(pDifferenceFile, blockNumber);
		}

		if (ans == SCPE_OK)
		{
			ans = diff_update_existing_block(pDifferenceFile, blockNumber, pBuffer, blockOffset, byteCount);
		}
	}

	return ans;
}

t_stat diff_update_existing_block(DIFFERENCE_FILE *pDifferenceFile, t_addr blockNumber, uint8 *pBuffer, uint32 blockOffset, size_t byteCount)
{
	t_stat ans = diff_check_block_number(pDifferenceFile, blockNumber);

	if (ans == SCPE_OK)
	{
		/*fprintf(sim_deb,"  Updating block %I64u, Offset %u, writing %u bytes\r\n", blockNumber, blockOffset, byteCount);*/
		ans = diff_move(pDifferenceFile->DifferenceFile, pDifferenceFile->pHeader->BlockIndex[blockNumber] + blockOffset);

		if (ans == SCPE_OK) ans = diff_direct_fwrite(pBuffer, sizeof(uint8), byteCount, pDifferenceFile->DifferenceFile);
	}

	return ans;
}

t_stat diff_create_new_block(DIFFERENCE_FILE *pDifferenceFile, t_addr blockNumber)
{
	t_stat ans = diff_check_block_number(pDifferenceFile, blockNumber);

	if (ans == SCPE_OK)
	{
		t_addr newBlockPosition;

		ans = diff_move_eof(pDifferenceFile->DifferenceFile);
		newBlockPosition = ftell(pDifferenceFile->DifferenceFile);
		/*fprintf(sim_deb,"Appending new block at position %I64u for block %I64u\r\n", newBlockPosition, blockNumber);*/
		if (ans == SCPE_OK)
		{
			ans = diff_read_original_block(pDifferenceFile, blockNumber);
		}
		
		if (ans == SCPE_OK)
		{
			ans = diff_direct_fwrite(pDifferenceFile->pHeader->Buffer, sizeof(uint8), pDifferenceFile->pHeader->BlockSize, pDifferenceFile->DifferenceFile);
		}

		if (ans == SCPE_OK)
		{
			ans = diff_update_block_index_entry(pDifferenceFile, blockNumber, newBlockPosition);
		}
	}

	return ans;
}

/* Assumes data will be flipped by the caller */
size_t diff_fread (void *bptr, size_t size, size_t count, FILE *fptr)
{
	static int fread_count = 0;
	size_t ans;
	size_t bytesRead = 0;
	t_stat stat = SCPE_OK;
	t_addr startAddress = ftell(fptr);
	DIFFERENCE_FILE *df;

	/* Read the whole block from the original disk so that the original disk file position is correct, then overlay with difference data */
	ans = fread(bptr, size, count, fptr);

	df = diff_find_difference_file_for_file(fptr);
	if (df != NULL)
	{
		uint32 currentBlockOffset = startAddress % df->pHeader->BlockSize; /* offset into the current block */
		size_t bufferLength = size * count;
		t_addr startBlock = diff_calculate_block_number_for_address(df, startAddress);
		t_addr endBlock = diff_calculate_block_number_for_address(df, startAddress + bufferLength - 1);
		t_addr i;

		/*fprintf(sim_deb,"Start buffer read at %I64u, length = %u\r\n", startAddress, bufferLength);*/

		for (i = startBlock; i <= endBlock && stat == SCPE_OK; i++)
		{
			size_t bytesToRead;
			bytesToRead = df->pHeader->BlockSize - currentBlockOffset % df->pHeader->BlockSize;
			if (bytesToRead > bufferLength - bytesRead)
			{
				bytesToRead = bufferLength - bytesRead;
			}
			
			stat = diff_read_block(df, i, (uint8 *)bptr + bytesRead, currentBlockOffset, bytesToRead);
			
			if (stat == SCPE_OK)
			{
				bytesRead += bytesToRead;
				currentBlockOffset = (startAddress + bytesRead) % df->pHeader->BlockSize;
			}
		}

		/*if (fread_count<60)
		{
			diff_trace_block(bptr, bufferLength, "");
			fread_count++;
		}*/
	}

	return ans;
}

t_stat diff_read_block(DIFFERENCE_FILE *pDifferenceFile, t_addr blockNumber, uint8 *pBuffer, uint32 blockOffset, uint32 byteCount)
{
	t_stat ans = diff_check_block_number(pDifferenceFile, blockNumber);

	/*fprintf(sim_deb,"  Diff block read. Block %I64u, offset = %u, length = %u\r\n", blockNumber, blockOffset, byteCount);*/
	if (ans == SCPE_OK)
	{
		if (pDifferenceFile->pHeader->BlockIndex[blockNumber] != 0)
		{
			ans = diff_read_diff_block(pDifferenceFile, blockNumber, pBuffer, blockOffset, byteCount);
		}
	}

	return ans;
}

t_stat diff_read_diff_block(DIFFERENCE_FILE *pDifferenceFile, t_addr blockNumber, uint8 *pBuffer, uint32 blockOffset, uint32 byteCount)
{
	t_stat ans = diff_check_block_number(pDifferenceFile, blockNumber);

	t_addr diffFileLocation = pDifferenceFile->pHeader->BlockIndex[blockNumber] + blockOffset;
	/*fprintf(sim_deb,"  Diff block read. Block %I64u, offset = %u, length = %u at address %I64u\r\n", blockNumber, blockOffset, byteCount, diffFileLocation);*/
	if (ans == SCPE_OK)
	{
		ans = diff_move(pDifferenceFile->DifferenceFile, diffFileLocation);
		if (ans == SCPE_OK)
		{
			ans = diff_direct_fread(pBuffer, 1, byteCount, pDifferenceFile->DifferenceFile);
		}
	}

	return ans;
}

t_stat diff_read_original_block(DIFFERENCE_FILE *pDifferenceFile, t_addr blockNumber)
{
	t_stat ans = SCPE_OK;

	t_addr originalPosition = ftell(pDifferenceFile->pUnit->fileref);

	ans = diff_move(pDifferenceFile->pUnit->fileref, pDifferenceFile->pHeader->BlockSize * blockNumber);
	if (ans == SCPE_OK)
	{
		ans = diff_direct_fread(pDifferenceFile->pHeader->Buffer, 1, pDifferenceFile->pHeader->BlockSize, pDifferenceFile->pUnit->fileref);
	}

	diff_move(pDifferenceFile->pUnit->fileref, originalPosition);

	return ans;
}

t_stat diff_move(FILE *fptr, t_addr position)
{
	t_stat ans = SCPE_OK;

	fseek(fptr, (size_t)position, SEEK_SET);  /* cast on position will cause failure on very large files, needs 64-bit target */
	if (ferror(fptr)) ans = SCPE_IOERR;

	return ans;
}

t_stat diff_move_eof(FILE *fptr)
{
	t_stat ans =SCPE_OK;

	fseek(fptr, 0, SEEK_END);
	if (ferror(fptr)) ans = SCPE_IOERR;

	return ans;
}

t_stat diff_check_block_number(DIFFERENCE_FILE *pDifferenceFile, t_addr blockNumber)
{
	t_stat ans = SCPE_OK;

	if (blockNumber < 0 || blockNumber >= pDifferenceFile->pHeader->NumBlocks)
	{
		printf("Invalid difference block number %I64u, must in range 0 to %I64u\n", blockNumber, pDifferenceFile->pHeader->NumBlocks - 1);
		ans = SCPE_ARG;
	}

	return ans;
}

void diff_show_stats(DIFFERENCE_FILE *pDifferenceFile)
{
	t_addr i;
	t_addr count = 0;

	for (i = 0; i < pDifferenceFile->pHeader->NumBlocks; i++)
	{
		if (pDifferenceFile->pHeader->BlockIndex[i] != 0)
		{
			/*fprintf(sim_deb, "Block %I64u in use, address is %I64u\r\n", i, pDifferenceFile->pHeader->BlockIndex[i]);*/
			count++;
		}
	}

	printf("Using %I64u blocks, out of a total %I64u, this is %.2f%%\n", count, pDifferenceFile->pHeader->NumBlocks, ((double)count * 100.0)/(double)pDifferenceFile->pHeader->NumBlocks);
	/*printf("Block index 0 value is %I64u\n", pDifferenceFile->pHeader->BlockIndex[0]);*/
}

t_stat diff_merge_difference_file(DIFFERENCE_FILE *df)
{
	t_stat ans = SCPE_OK;
	t_addr i;

	for (i = 0; i < df->pHeader->NumBlocks && ans == SCPE_OK; i++)
	{
		if (df->pHeader->BlockIndex[i] != 0)
		{
			ans = diff_read_diff_block(df, i, df->pHeader->Buffer, 0, df->pHeader->BlockSize);
			if (ans == SCPE_OK) ans = diff_move(df->pUnit->fileref, i * df->pHeader->BlockSize);
			if (ans == SCPE_OK) ans = diff_direct_fwrite(df->pHeader->Buffer, 1, df->pHeader->BlockSize, df->pUnit->fileref); 
		}
	}

	return ans;
}

t_stat diff_indirect_fwrite(void *bptr, size_t size, size_t count, FILE *fptr)
{
	t_stat ans = SCPE_OK;
	sim_fwrite(bptr, size, count, fptr);
	if (ferror(fptr)) ans = SCPE_IOERR;
	return ans;
}

t_stat diff_indirect_fread(void *bptr, size_t size, size_t count, FILE *fptr)
{
	t_stat ans = SCPE_OK;
	sim_fread(bptr, size, count, fptr);
	if (ferror(fptr)) ans = SCPE_IOERR;
	return ans;
}

t_stat diff_direct_fwrite(void *bptr, size_t size, size_t count, FILE *fptr)
{
	t_stat ans = SCPE_OK;
	fwrite(bptr, size, count, fptr);
	if (ferror(fptr)) ans = SCPE_IOERR;
	return ans;
}

t_stat diff_direct_fread(void *bptr, size_t size, size_t count, FILE *fptr)
{
	t_stat ans = SCPE_OK;
	fread(bptr, size, count, fptr);
	if (ferror(fptr)) ans = SCPE_IOERR;
	return ans;
}

t_stat diff_fflush(FILE *fptr)
{
	t_stat ans = SCPE_OK;
	fflush(fptr);
	if (ferror(fptr)) ans = SCPE_IOERR;
	return ans;
}

t_stat sim_diff_attach (DEVICE *dptr, UNIT *uptr, char *fileName, t_value blockSize)
{
	t_stat ans = SCPE_OK;

	DIFFERENCE_FILE *df = diff_find_difference_file_for_unit(uptr);
	if (df != NULL)
	{
		ans = SCPE_ALATT;
	}
	else
	{
		ans = diff_allocate_difference_file_table_slot(uptr, &df);
		if (ans == SCPE_OK)
		{
			df->FileName = fileName;
			ans = diff_open_difference_file(df, (uint32)blockSize);
		}
	}

	return ans;
}

t_stat sim_diff_detach (DEVICE *dptr, UNIT *uptr)
{
	t_stat ans = SCPE_OK;

	DIFFERENCE_FILE *df = diff_find_difference_file_for_unit(uptr);
	if (df == NULL)
	{
		ans = SCPE_NOATTD;
	}
	else
	{
		diff_deallocate_difference_file_table_slot(df);
		diff_close_difference_file(df);
	}

	return ans;
}

t_stat sim_diff_compare (DEVICE *dptr, UNIT *uptr, char *virtualFileName, char *diffFileName, t_value blockSize)
{
	t_stat ans = SCPE_OK;
	DIFFERENCE_FILE df;
	t_addr comparandFileSize;
	FILE *comparandFile;

	df.FileName = diffFileName;
	df.pUnit = uptr;

	comparandFile = sim_fopen(virtualFileName, "rb"); /* read only */
	if (comparandFile != NULL)
	{
		diff_move_eof(comparandFile);
		comparandFileSize = ftell(comparandFile);
		if (comparandFileSize > uptr->capac)
		{
			printf("File %s is bigger than the virtual disk being compared to so cannot be compared\n", virtualFileName);
			ans = SCPE_ARG;
		}
	}
	else
	{
		ans = SCPE_OPENERR;
	}

	if (ans == SCPE_OK)
	{
		int i;
		void *comparandBuffer = malloc((size_t)blockSize);

		ans = diff_create_difference_file(&df, diffFileName, (uint32)blockSize);

		diff_move(uptr->fileref, 0);
		diff_move(comparandFile, 0);
		for (i = 0; i<df.pHeader->NumBlocks && ans == SCPE_OK; i++)
		{
			ans = diff_direct_fread(comparandBuffer, 1, (size_t)blockSize, comparandFile);
			if (ans == SCPE_OK) ans = diff_direct_fread(df.pHeader->Buffer, 1, (size_t)blockSize, uptr->fileref);

			if (ans == SCPE_OK)
			{
				if (memcmp(comparandBuffer, df.pHeader->Buffer, (size_t)blockSize)!=0)
				{
					ans = diff_create_new_block(&df, i);
					if (ans == SCPE_OK)
					{
						ans = diff_update_existing_block(&df, i, comparandBuffer, 0, (size_t)blockSize);
					}
				}
			}
		}
			
		if (ans == SCPE_OK)
		{
			diff_show_stats(&df);
			diff_close_difference_file(&df);
		}

		free(comparandBuffer);
	}

	fclose(comparandFile);

	return ans;
}

t_stat sim_diff_merge (DEVICE *dptr, UNIT *uptr)
{
	t_stat ans = SCPE_OK;

	DIFFERENCE_FILE *df = diff_find_difference_file_for_unit(uptr);
	if (df == NULL)
	{
		ans = SCPE_NOATTD;
	}
	else
	{
		ans = diff_merge_difference_file(df);
		if (ans == SCPE_OK)
		{
			diff_deallocate_difference_file_table_slot(df);
			diff_close_difference_file(df);
			diff_delete_difference_file(df);
		}
	}

	return ans;
}

t_stat sim_diff_stats (DEVICE *dptr, UNIT *uptr)
{
	int i;
	t_stat ans = SCPE_OK;

	for (i = 0; i < MAX_DIFFERENCE_FILES; i++)
	{
		DIFFERENCE_FILE *pdf = &DifferenceFileTable[i];
		if (pdf->InUse)
		{
			diff_show_stats(pdf);
		}
	}

	return ans;
}
